@@ -39,16 +39,7 @@
 
 #include "hfpmodem.h"
 
-#define AG_CHLD_0 0x01
-#define AG_CHLD_1 0x02
-#define AG_CHLD_1x 0x04
-#define AG_CHLD_2 0x08
-#define AG_CHLD_2x 0x10
-#define AG_CHLD_3 0x20
-#define AG_CHLD_4 0x40
-
 static const char *none_prefix[] = { NULL };
-static const char *chld_prefix[] = { "+CHLD:", NULL };
 static const char *clcc_prefix[] = { "+CLCC:", NULL };
 
 struct voicecall_data {
@@ -58,7 +49,6 @@ struct voicecall_data {
 	unsigned int ag_mpty_features;
 	unsigned char cind_pos[HFP_INDICATOR_LAST];
 	int cind_val[HFP_INDICATOR_LAST];
-	unsigned int id_list;
 	unsigned int local_release;
 };
 
@@ -76,10 +66,26 @@ struct change_state_req {
 	int affected_types;
 };
 
-static struct ofono_call *create_call(struct voicecall_data *d, int type,
+static GSList *find_dialing(GSList *calls)
+{
+	GSList *c;
+
+	c = g_slist_find_custom(calls, GINT_TO_POINTER(CALL_STATUS_DIALING),
+				at_util_call_compare_by_status);
+
+	if (!c)
+		c = g_slist_find_custom(calls,
+					GINT_TO_POINTER(CALL_STATUS_ALERTING),
+					at_util_call_compare_by_status);
+
+	return c;
+}
+
+static struct ofono_call *create_call(struct ofono_voicecall *vc, int type,
 					int direction, int status,
 					const char *num, int num_type, int clip)
 {
+	struct voicecall_data *d = ofono_voicecall_get_data(vc);
 	struct ofono_call *call;
 
 	/* Generate a call structure for the waiting call */
@@ -88,7 +94,7 @@ static struct ofono_call *create_call(struct voicecall_data *d, int type,
 	if (!call)
 		return NULL;
 
-	call->id = at_util_alloc_next_id(&d->id_list);
+	call->id = ofono_voicecall_get_next_callid(vc);
 	call->type = type;
 	call->direction = direction;
 	call->status = status;
@@ -110,10 +116,9 @@ static struct ofono_call *new_call_notify(struct ofono_voicecall *vc, int type,
 					int direction, int status,
 					const char *num, int num_type, int clip)
 {
-	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
 	struct ofono_call *c;
 
-	c = create_call(vd, type, direction, status, num, num_type, clip);
+	c = create_call(vc, type, direction, status, num, num_type, clip);
 
 	ofono_voicecall_notify(vc, c);
 
@@ -134,7 +139,6 @@ static void release_call(struct ofono_voicecall *vc, struct ofono_call *call)
 		reason = OFONO_DISCONNECT_REASON_REMOTE_HANGUP;
 
 	ofono_voicecall_disconnected(vc, call->id, reason, NULL);
-	at_util_release_id(&vd->id_list, call->id);
 	vd->local_release &= ~(1 << call->id);
 
 	g_free(call);
@@ -161,6 +165,7 @@ static void release_with_status(struct ofono_voicecall *vc, int status)
 	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
 	GSList *p = NULL;
 	GSList *c = vd->calls;
+	GSList *t;
 	struct ofono_call *call;
 
 	while (c) {
@@ -179,7 +184,9 @@ static void release_with_status(struct ofono_voicecall *vc, int status)
 		else
 			vd->calls = c->next;
 
-			g_slist_free_1(c);
+		t = c;
+		c = c->next;
+		g_slist_free_1(t);
 	}
 }
 
@@ -217,7 +224,6 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 				ofono_voicecall_disconnected(vc, oc->id,
 								reason, NULL);
 
-			at_util_release_id(&vd->id_list, oc->id);
 			vd->local_release &= ~(1 << oc->id);
 
 			o = o->next;
@@ -301,11 +307,11 @@ static void atd_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		if (call->status != 0)
 			continue;
 
-		call->status = 2;
+		call->status = CALL_STATUS_HELD;
 		ofono_voicecall_notify(vc, call);
 	}
 
-	call = create_call(vd, 0, 0, CALL_STATUS_DIALING, NULL, type, validity);
+	call = create_call(vc, 0, 0, CALL_STATUS_DIALING, NULL, type, validity);
 
 	if (!call) {
 		ofono_error("Unable to allocate call, "
@@ -446,6 +452,45 @@ static void hfp_release_all_active(struct ofono_voicecall *vc,
 	CALLBACK_WITH_FAILURE(cb, data);
 }
 
+static void hfp_send_dtmf(struct ofono_voicecall *vc, const char *dtmf,
+			ofono_voicecall_cb_t cb, void *data)
+{
+	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
+	struct change_state_req *req = g_try_new0(struct change_state_req, 1);
+	char *buf;
+	int s;
+
+	if (!req)
+		goto error;
+
+	req->vc = vc;
+	req->cb = cb;
+	req->data = data;
+	req->affected_types = 0;
+
+	/* strlen("AT+VTS=") = 7 */
+	buf = g_try_new(char, strlen(dtmf) + 7);
+
+	if (!buf)
+		goto error;
+
+	sprintf(buf, "AT+VTS=%s", dtmf);
+
+	s = g_at_chat_send(vd->chat, buf, none_prefix,
+				generic_cb, req, g_free);
+
+	g_free(buf);
+
+	if (s > 0)
+		return;
+
+error:
+	if (req)
+		g_free(req);
+
+	CALLBACK_WITH_FAILURE(cb, data);
+}
+
 static void no_carrier_notify(GAtResult *result, gpointer user_data)
 {
 	DBG("");
@@ -462,6 +507,12 @@ static void ccwa_notify(GAtResult *result, gpointer user_data)
 
 	dump_response("ccwa_notify", TRUE, result);
 
+	/* CCWA can repeat, ignore if we already have an waiting call */
+	if (g_slist_find_custom(vd->calls,
+				GINT_TO_POINTER(CALL_STATUS_WAITING),
+				at_util_call_compare_by_status))
+		return;
+
 	g_at_result_iter_init(&iter, result);
 
 	if (!g_at_result_iter_next(&iter, "+CCWA:"))
@@ -480,7 +531,7 @@ static void ccwa_notify(GAtResult *result, gpointer user_data)
 
 	ofono_debug("ccwa_notify: %s %d %d", num, num_type, validity);
 
-	call = create_call(vd, 0, 1, 5, num, num_type, validity);
+	call = create_call(vc, 0, 1, 5, num, num_type, validity);
 
 	if (!call) {
 		ofono_error("malloc call structfailed. Call management is fubar");
@@ -529,7 +580,7 @@ static void ring_notify(GAtResult *result, gpointer user_data)
 	}
 
 	/* Generate an incoming call of voice type */
-	call = create_call(vd, 0, 1, CALL_STATUS_INCOMING, NULL, 128, 2);
+	call = create_call(vc, 0, 1, CALL_STATUS_INCOMING, NULL, 128, 2);
 
 	if (!call)
 		ofono_error("Couldn't create call, call management is fubar!");
@@ -599,36 +650,25 @@ static void ciev_call_notify(struct ofono_voicecall *vc,
 
 	switch (value) {
 	case 0:
-	{
-		GSList *waiting;
-		GSList *incoming;
-
 		/* If call goes to 0, then we have no held or active calls
 		 * in the system.  The waiting calls are promoted to incoming
-		 * calls
+		 * calls, dialing calls are kept.  This also handles the
+		 * situation when dialing and waiting calls exist
 		 */
-		waiting = g_slist_find_custom(vd->calls,
-					GINT_TO_POINTER(CALL_STATUS_WAITING),
-					at_util_call_compare_by_status);
+		release_with_status(vc, CALL_STATUS_HELD);
+		release_with_status(vc, CALL_STATUS_ACTIVE);
 
-		if (waiting) {
-			incoming = waiting;
-			call = waiting->data;
-			call->status = CALL_STATUS_INCOMING;
-			ofono_voicecall_notify(vc, call);
-		} else
-			incoming = g_slist_find_custom(vd->calls,
-					GINT_TO_POINTER(CALL_STATUS_INCOMING),
-					at_util_call_compare_by_status);
+		/* Promote waiting to incoming if it is the last call */
+		if (vd->calls && vd->calls->next == NULL) {
+			call = vd->calls->data;
 
-		if (incoming)
-			vd->calls = g_slist_remove_link(vd->calls, incoming);
-
-		release_all_calls(vc);
-		vd->calls = incoming;
+			if (call->status == CALL_STATUS_WAITING) {
+				call->status = CALL_STATUS_INCOMING;
+				ofono_voicecall_notify(vc, call);
+			}
+		}
 
 		break;
-	}
 
 	case 1:
 	{
@@ -681,24 +721,12 @@ static void sync_dialing_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		return;
 
 	/* Look for dialing or alerting calls on the new list */
-	n = g_slist_find_custom(calls, GINT_TO_POINTER(CALL_STATUS_DIALING),
-				at_util_call_compare_by_status);
-
-	if (!n)
-		n = g_slist_find_custom(calls,
-					GINT_TO_POINTER(CALL_STATUS_ALERTING),
-					at_util_call_compare_by_status);
+	n = find_dialing(calls);
 
  	/* Let us find if we have done the dial from HF by looking for
 	 * existing dialing or alerting calls
 	 */
-	o = g_slist_find_custom(vd->calls, GINT_TO_POINTER(CALL_STATUS_DIALING),
-				at_util_call_compare_by_status);
-
-	if (!o)
-		o = g_slist_find_custom(vd->calls,
-					GINT_TO_POINTER(CALL_STATUS_ALERTING),
-					at_util_call_compare_by_status);
+	o = find_dialing(vd->calls);
 
 	if (!n && o) {
 		oc = o->data;
@@ -733,14 +761,7 @@ static void ciev_callsetup_notify(struct ofono_voicecall *vc,
 	GSList *dialing;
 	GSList *waiting;
 
-	dialing = g_slist_find_custom(vd->calls,
-					GINT_TO_POINTER(CALL_STATUS_DIALING),
-					at_util_call_compare_by_status);
-
-	if (!dialing)
-		dialing = g_slist_find_custom(vd->calls,
-					GINT_TO_POINTER(CALL_STATUS_ALERTING),
-					at_util_call_compare_by_status);
+	dialing = find_dialing(vd->calls);
 
 	waiting = g_slist_find_custom(vd->calls,
 					GINT_TO_POINTER(CALL_STATUS_WAITING),
@@ -767,6 +788,12 @@ static void ciev_callsetup_notify(struct ofono_voicecall *vc,
 			goto out;
 		}
 
+		/* If call=1 and no call is waiting or dialing, the call is
+		 * active and we moved it to active state when call=1 arrived
+		 */
+		if (waiting == NULL && dialing == NULL)
+			goto out;
+
 		 /*
 		 * If call=1, in the waiting case we have to poll, since we
 		 * have no idea whether a waiting call gave up or we accepted
@@ -847,38 +874,13 @@ static void ciev_callheld_notify(struct ofono_voicecall *vc,
 		break;
 
 	case 1:
-	{
-		GSList *waiting;
-
-		waiting = g_slist_find_custom(vd->calls,
-					GINT_TO_POINTER(CALL_STATUS_WAITING),
-					at_util_call_compare_by_status);
-
-		for (l = vd->calls; l; l = l->next) {
-			call = l->data;
-
-			if (waiting) {
-				if (call->status == CALL_STATUS_WAITING) {
-					call->status = CALL_STATUS_ACTIVE;
-					ofono_voicecall_notify(vc, call);
-				} else if (call->status == CALL_STATUS_ACTIVE) {
-					call->status = CALL_STATUS_HELD;
-					ofono_voicecall_notify(vc, call);
-				}
-			} else {
-				if (call->status == CALL_STATUS_ACTIVE) {
-					call->status = CALL_STATUS_HELD;
-					ofono_voicecall_notify(vc, call);
-				} else if (call->status == CALL_STATUS_HELD) {
-					call->status = CALL_STATUS_ACTIVE;
-					ofono_voicecall_notify(vc, call);
-				}
-			}
-		}
-
+		/* We have to poll here, we have no idea whether the call was
+		 * accepted by CHLD=1 or swapped by CHLD=2 or one call was
+		 * chosed for private chat by CHLD=2x
+		 */
+		g_at_chat_send(vd->chat, "AT+CLCC", clcc_prefix,
+				clcc_poll_cb, vc, NULL);
 		break;
-	}
-
 	case 2:
 		if (callheld == 0) {
 			for (l = vd->calls; l; l = l->next) {
@@ -925,45 +927,19 @@ static void ciev_notify(GAtResult *result, gpointer user_data)
 		ciev_callheld_notify(vc, value);
 }
 
-static void chld_cb(gboolean ok, GAtResult *result, gpointer user_data)
+static void hfp_clcc_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
-	struct voicecall_data *vd = user_data;
-	unsigned int ag_mpty_feature = 0;
-	GAtResultIter iter;
-	const char *str;
+	struct ofono_voicecall *vc = user_data;
+	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
+	GSList *l;
 
 	if (!ok)
 		return;
 
-	g_at_result_iter_init(&iter, result);
+	vd->calls = at_util_parse_clcc(result);
 
-	if (!g_at_result_iter_next(&iter, "+CHLD:"))
-		return;
-
-	if (!g_at_result_iter_open_list(&iter))
-		return;
-
-	while (g_at_result_iter_next_unquoted_string(&iter, &str)) {
-		if (!strcmp(str, "0"))
-			ag_mpty_feature |= AG_CHLD_0;
-		else if (!strcmp(str, "1"))
-			ag_mpty_feature |= AG_CHLD_1;
-		else if (!strcmp(str, "1x"))
-			ag_mpty_feature |= AG_CHLD_1x;
-		else if (!strcmp(str, "2"))
-			ag_mpty_feature |= AG_CHLD_2;
-		else if (!strcmp(str, "2x"))
-			ag_mpty_feature |= AG_CHLD_2x;
-		else if (!strcmp(str, "3"))
-			ag_mpty_feature |= AG_CHLD_3;
-		else if (!strcmp(str, "4"))
-			ag_mpty_feature |= AG_CHLD_4;
-	}
-
-	if (!g_at_result_iter_close_list(&iter))
-		return;
-
-	vd->ag_mpty_features = ag_mpty_feature;
+	for (l = vd->calls; l; l = l->next)
+		ofono_voicecall_notify(vc, l->data);
 }
 
 static void hfp_voicecall_initialized(gboolean ok, GAtResult *result,
@@ -983,6 +959,9 @@ static void hfp_voicecall_initialized(gboolean ok, GAtResult *result,
 				no_carrier_notify, FALSE, vc, NULL);
 
 	ofono_voicecall_register(vc);
+
+	/* Populate the call list */
+	g_at_chat_send(vd->chat, "AT+CLCC", clcc_prefix, hfp_clcc_cb, vc, NULL);
 }
 
 static int hfp_voicecall_probe(struct ofono_voicecall *vc, unsigned int vendor,
@@ -995,14 +974,11 @@ static int hfp_voicecall_probe(struct ofono_voicecall *vc, unsigned int vendor,
 
 	vd->chat = data->chat;
 	vd->ag_features = data->ag_features;
+	vd->ag_mpty_features = data->ag_mpty_features;
 
 	memcpy(vd->cind_pos, data->cind_pos, HFP_INDICATOR_LAST);
 	memcpy(vd->cind_val, data->cind_val, HFP_INDICATOR_LAST);
 
-	if (vd->ag_features & AG_FEATURE_3WAY)
-		g_at_chat_send(vd->chat, "AT+CHLD=?", chld_prefix,
-			chld_cb, vd, NULL);
-
 	ofono_voicecall_set_data(vc, vd);
 
 	g_at_chat_send(vd->chat, "AT+CLIP=1", NULL, NULL, NULL, NULL);
@@ -1030,7 +1006,6 @@ static struct ofono_voicecall_driver driver = {
 	.dial			= hfp_dial,
 	.answer			= hfp_answer,
 	.hangup			= hfp_hangup,
-	.list_calls		= NULL,
 	.hold_all_active	= hfp_hold_all_active,
 	.release_all_held	= hfp_release_all_held,
 	.set_udub		= hfp_set_udub,
@@ -1041,7 +1016,7 @@ static struct ofono_voicecall_driver driver = {
 	.transfer		= NULL,
 	.deflect		= NULL,
 	.swap_without_accept	= NULL,
-	.send_tones		= NULL
+	.send_tones		= hfp_send_dtmf
 };
 
 void hfp_voicecall_init()
