@@ -34,3 +34,8 @@ ssize_t read_file(unsigned char *buffer, size_t len,
 
 ssize_t write_file(const unsigned char *buffer, size_t len, mode_t mode,
 			const char *path_fmt, ...);
+
+GKeyFile *storage_open(const char *imsi, const char *store);
+void storage_sync(const char *imsi, const char *store, GKeyFile *keyfile);
+void storage_close(const char *imsi, const char *store, GKeyFile *keyfile,
+			gboolean save);
